/****************************************************************************\
  FILE..........: Policy.h
  AUTHOR........: Kari Laasonen
  COPYRIGHT.....: Copyright (c) 1998, Data Fellows Ltd.
  VERSION.......: 1.00
  DESCRIPTION...: Policy API/Configuration handler for user mode components

  HISTORY.......: DATE     COMMENT
                  -------- ------------------------------------------------
                  26-May-98 Created.
				  17-Aug-98 Modifications to function prototypes
				  02-Sep-98 Change notification functionality
				  01-Oct-98 A few error codes added
				  21-Oct-98 DfpInitialize, access rights, and more error codes
\****************************************************************************/

#ifndef POLICY_H
#define POLICY_H

#include <stdlib.h>

#ifndef FSMADEF_H
#include "fsmadef.h"
#endif

#ifndef POLCYDEF_H
#include "polcydef.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _MSC_VER  /* We're using Visual C++ compiler */
#  ifndef KERNEL_MODE
#    ifdef POLICY_DLL
#      define DFP_EXPORT __declspec(dllexport)
#    else
#      define DFP_EXPORT __declspec(dllimport)
#    endif
#  else /* KERNEL_MODE */
#    define DFP_EXPORT
#  endif
#  define DFP_CALLMETHOD __stdcall
#else /* _MSC_VER */
#  define DFP_EXPORT
#  define DFP_CALLMETHOD
#endif

#define DFP_STDCALL DFP_EXPORT DfPolicyStatus DFP_CALLMETHOD
#define DFP_LIBCALL DfPolicyStatus DFP_CALLMETHOD
#define DFP_CALL(T) DFP_EXPORT T DFP_CALLMETHOD

/* Variable types */
enum DfPolicyType {
    DFP_UNDEFINED,
    DFP_NULL,
    DFP_OID,
    DFP_STRING,
    DFP_DISPLAYSTRING,
	DFP_ADDRESS,
    DFP_OPAQUE,
    DFP_NUMBER,
    DFP_COUNTER,
    DFP_GAUGE,
    DFP_TICKS,
    DFP_TABLE
};

#define DFP_MIN_SCALAR_TYPE DFP_UNDEFINED
#define DFP_MAX_SCALAR_TYPE DFP_TICKS

typedef struct {
    int length;      /* Number of characters in the string */
    char buffer[1];  /* Character array */
    /* Warning: this is a variable-length structure */
} DfPolicyString;

/* Integer SNMP variables */
typedef unsigned int df_uint;
typedef int df_int;

typedef struct {
    enum DfPolicyType type;
    union {
		DfPolicyString oid;
		DfPolicyString string;
		DfPolicyString dstring; /* DisplayString */
		DfPolicyString opaque;
		DfPolicyString address;
		df_uint counter;
		df_uint gauge;
		df_uint ticks;
		df_int number;
    } data;
} DfPolicyValue;

/* Context used with DfpGetNextChangedOID */
typedef struct {
	FSMAMESSAGE message;
	df_uint count;		/* Number of changed OIDs in message */
	df_uint serial;		/* Serial number */
	void* next;			/* Opaque information used by library */
} DfNotifyContext;

/* Sources */
enum DfPolicySource {
	DFP_SRC_UNKNOWN,
	DFP_SRC_DEFAULT,
	DFP_SRC_BASE,
	DFP_SRC_INCREMENTAL
};

/* Access rights */
enum DfPolicyAccess {
	DFP_NO_ACCESS = 0,
	DFP_READ_ACCESS = 1,
	DFP_WRITE_ACCESS = 2,
	DFP_ALL_ACCESS = DFP_READ_ACCESS | DFP_WRITE_ACCESS
};

/* Most routines return zero on success, or a negative error code on failure */
typedef long DfPolicyStatus;

#define DFP_SUCCESS				0

#define DFP_ERR_INVALID         -1	/* DfPolicyVar is not valid */
#define DFP_ERR_BUFFER          -2	/* Insufficient buffer space provided */
#define DFP_ERR_TYPE            -3	/* Bad type field, or table cell type differs from column type */
#define DFP_ERR_MEM             -4	/* Memory allocation/reallocation failed */
#define DFP_ERR_INDEX           -5	/* Bad row/column/choice index */
#define DFP_ERR_TABLE           -6	/* Table operations cannot be used on scalar variables or vice versa */
#define DFP_ERR_OID_SYNTAX      -7  /* Set OID failed: OID string is bad */
#define DFP_ERR_PARAM           -8	/* Function parameter out of valid range */
#define DFP_ERR_NO_DATA         -9	/* PolicyVar doesn't have the requested data */
#define DFP_ERR_CONVERSION      -10 /* UTF8 string conversion failed */
#define DFP_ERR_NO_DLL			-11 /* Required DLL not found */

#define DFP_ERR_WRONG_TYPE      -511 /* Set result: variable type conflicts with existing policy */
#define DFP_ERR_BAD_TABLE       -510 /* Set result: your table contains multiple identical rows etc. */
#define DFP_ERR_RESOURCE        -509 /* Set result: driver encountered resource shortage */
#define DFP_ERR_NO_SUCH_OID     -508 /* Get result: OID doesn't exist */
#define DFP_ERR_NO_VALUE        -507 /* Get result: returned variable is undefined */
#define DFP_ERR_FINAL           -506 /* Set result: base policy variable is final */
#define DFP_ERR_RESTRICTION     -505 /* Set result: set not allowed by base policy restrictions (range/choice) */
#define DFP_ERR_DRIVER_PARAM    -504 /* ioctl() parameters were not accepted by driver */
#define DFP_ERR_DRIVER_ERROR    -503 /* Get/set result: driver communications failed */
#define DFP_ERR_NO_DRIVER       -502 /* Get/set result: no policy driver found */
#define DFP_ERR_DRIVER_VERSION  -501 /* Get/set result: driver and FSPMAPI versions don't match */
#define DFP_ERR_CORRUPTED_FILE  -500 /* Get/set result: (inc) policy is corrupted */

#define DFP_ERR_FILE_ERROR      -480 /* Policy driver: cannot read policy file contents */
#define DFP_ERR_DER_PARSING     -479 /* PD: signature/public key parsing */
#define DFP_ERR_KEY_ALGORITHM   -478 /* PD: unknown signature/public key algorithm */
#define DFP_ERR_NO_PUBLIC_KEY   -477 /* PD: driver doesn't have public key for signature checking */
#define DFP_ERR_BAD_SIGNATURE   -476 /* PD: file signature checking fails */
#define DFP_ERR_NO_FILE         -475 /* PD: policy file doesn't exist */
#define DFP_ERR_FILE_ID         -474 /* future */
#define DFP_ERR_FILE_VERSION    -473 /* PD: file version is not known */
#define DFP_ERR_PARSE_BAD_CHAR  -472 /* PD: parsing failed: unexpected character */
#define DFP_ERR_PARSE_EOF		-471 /* PD: parsing failed: unexpected EOF */
#define DFP_ERR_PARSE_NESTING	-470 /* PD: parsing failed: nesting too deep */

#define DFP_IS_PARSE_ERROR(x) ((x) >= DFP_ERR_PARSE_NESTING && \
							   (x) <= DFP_ERR_FILE_ID)

/*------------------------------------------------------------------- 
  Policy services, provided by Configuration Handler client library
  -------------------------------------------------------------------*/

/* This function must be called before any other Dfp* function call */
DFP_LIBCALL
DfpInitialize (FSMASESSION session);

/* Given OID, create a new policy variable object containing the value. */
DFP_LIBCALL
DfpGetPolicy (FSMASESSION session, const char* oid_string, DfPolicyVar* pvar);

/* As previous, but takes an OID variable instead of string. */
DFP_LIBCALL
DfpGetPolicyByOID (FSMASESSION session, DfPolicyOID oid, DfPolicyVar* pvar);

/* Set policy from a policy variable object which has an OID and a value. */
DFP_LIBCALL
DfpSetPolicy (FSMASESSION session, DfPolicyVar var);

/* Simple utility functions to get/set an integer value */
DFP_LIBCALL
DfpGetPolicyInt (FSMASESSION session,
				 const char* oid_string, df_uint* value, df_uint def_value);
DFP_LIBCALL
DfpSetPolicyInt (FSMASESSION session, 
				 const char* oid_string, df_uint value);

/* The following query any string/OID variable */
DFP_LIBCALL
DfpGetPolicyStringA (FSMASESSION session, const char* oid_string, 
					 char* value, size_t value_size, const char* def_value);
DFP_LIBCALL
DfpGetPolicyStringW (FSMASESSION session, const char* oid_string, 
					 wchar_t* value, size_t value_size, const wchar_t* def_value);
#ifdef UNICODE
#define DfpGetPolicyString DfpGetPolicyStringW
#else
#define DfpGetPolicyString DfpGetPolicyStringA
#endif /* UNICODE */

/* The following wrap DfpSetPolicyValue() */
DFP_LIBCALL
DfpSetPolicyStringA (FSMASESSION session, const char* oid_string, const char* value);
DFP_LIBCALL
DfpSetPolicyStringW (FSMASESSION session, const char* oid_string, const wchar_t* value);

#ifdef UNICODE
#define DfpSetPolicyString DfpSetPolicyStringW
#else
#define DfpSetPolicyString DfpSetPolicyStringA
#endif /* UNICODE */

/* Binary string is one that isn't transformed in any way */
DFP_LIBCALL
DfpGetPolicyBinary (FSMASESSION session, const char* oid_string, 
					unsigned char* value, size_t value_size, size_t* result_size);
DFP_LIBCALL
DfpSetPolicyBinary (FSMASESSION session, const char* oid_string, 
					const unsigned char* value, size_t value_size);

/*------------------------------------------------------------------- 
  Driver operations
  -------------------------------------------------------------------*/

DFP_LIBCALL
DfpRegisterOID (FSMASESSION session, DfPolicyOID pattern);

DFP_LIBCALL
DfpUnregister (FSMASESSION session);

DFP_LIBCALL
DfpGetFirstChange (FSMAMESSAGE msg, DfNotifyContext* context);

DFP_LIBCALL
DfpGetNextChange (DfNotifyContext* context, DfPolicyOID oid);

DFP_LIBCALL
DfpGetPublicKey (FSMASESSION session, void* data, size_t data_size, 
				 size_t* result_size);

/*------------------------------------------------------------------- 
  Operations on DfPolicyVar (and -OID) objects
  -------------------------------------------------------------------*/
DFP_STDCALL 
DfpCreatePolicyVar(DfPolicyVar* pvar);

DFP_STDCALL
DfpCreatePolicyTable(DfPolicyVar* pvar);

DFP_STDCALL 
DfpCreatePolicyOID(const char* oid_string, DfPolicyOID* poid);

DFP_STDCALL
DfpCopy (DfPolicyVar source, DfPolicyVar* dest_ptr);

DFP_STDCALL
DfpClose (DfPolicyVar var);

/* Packing and unpacking */
DFP_STDCALL
DfpCreatePackage (DfPolicyVar var, void** buffer, size_t* size);

DFP_STDCALL
DfpFreePackage (DfPolicyVar var, void* buffer, size_t size);

DFP_STDCALL
DfpCreatePolicyVarFromPackage (const void* buffer, size_t buffer_size, DfPolicyVar* pvar);

/* Comparisons */
DFP_CALL(int)
DfpIsEqual (DfPolicyVar var1, DfPolicyVar var2);

DFP_CALL(int)
DfpIsEqualValue (DfPolicyVar var1, DfPolicyVar var2);

DFP_STDCALL
DfpGetSource (DfPolicyVar var, enum DfPolicySource* src);

/*------------------------------------------------------------------- 
  OID operations
  -------------------------------------------------------------------*/

/* Return values (in addition to errors) from DfpCompareOID() */
#define DFP_LESS_THAN		1		/* OID 1 is less than OID 2 */
#define DFP_EQUAL			2		/* OID 1 is equal to OID 2 */
#define DFP_GREATER_THAN	3		/* OID 1 is greater than OID 2 */

DFP_CALL(int)
DfpCompareOID (DfPolicyVar var1, DfPolicyVar var2);

DFP_STDCALL
DfpGetOID (DfPolicyVar var, char* buffer, size_t buffer_size);

DFP_STDCALL
DfpSetOID (DfPolicyVar var, const char* oid_string);

DFP_STDCALL
DfpCopyOID (DfPolicyVar var, DfPolicyVar src);

DFP_STDCALL
DfpAppendOID (DfPolicyVar var, const char* oid_string);

DFP_STDCALL
DfpSetPartOID (DfPolicyVar var, int index, unsigned int part_value);

DFP_STDCALL
DfpGetDepthOID (DfPolicyVar var, int* depth);

DFP_CALL(int)
DfpCompleteOID (DfPolicyVar pattern_dest, DfPolicyVar source);

/*------------------------------------------------------------------- 
  Access to scalar variables
  -------------------------------------------------------------------*/
DFP_STDCALL
DfpGetType (DfPolicyVar var, enum DfPolicyType*);

DFP_STDCALL
DfpGetAccessRights (DfPolicyVar var, enum DfPolicyAccess*);

DFP_STDCALL
DfpGetValue (DfPolicyVar var, DfPolicyValue* value, size_t value_size);

DFP_STDCALL
DfpGetValueSize (DfPolicyVar var, size_t* value_size);

DFP_STDCALL
DfpGetInt (DfPolicyVar var, df_uint* value);

DFP_STDCALL
DfpGetStringA (DfPolicyVar var, char* value, size_t value_size);

DFP_STDCALL
DfpGetStringW (DfPolicyVar var, wchar_t* value, size_t value_size);

#ifdef UNICODE
#define DfpGetString DfpGetStringW
#else
#define DfpGetString DfpGetStringA
#endif /* UNICODE */

DFP_STDCALL
DfpGetBinary (DfPolicyVar var, 
			  unsigned char* value, size_t value_size, size_t* result_size);

DFP_STDCALL
DfpSetValue (DfPolicyVar var, const DfPolicyValue* value);

DFP_STDCALL
DfpSetInt (DfPolicyVar var, df_uint value);

DFP_STDCALL
DfpSetStringA (DfPolicyVar var, const char* value);

DFP_STDCALL
DfpSetStringW (DfPolicyVar var, const wchar_t* value);

#ifdef UNICODE
#define DfpSetString DfpSetStringW
#else
#define DfpSetString DfpSetStringA
#endif /* UNICODE */

DFP_STDCALL
DfpSetBinary (DfPolicyVar var, 
			  const unsigned char* value, size_t value_size);

DFP_STDCALL
DfpFormatStringA (DfPolicyVar var, char* buffer, size_t buffer_size);

DFP_STDCALL
DfpFormatStringW (DfPolicyVar var, wchar_t* buffer, size_t buffer_size);

#ifdef UNICODE
#define DfpFormatString DfpFormatStringW
#else
#define DfpFormatString DfpFormatStringA
#endif /* UNICODE */

DFP_STDCALL
DfpFormatTypeNameA (DfPolicyVar var, char* buffer, size_t buffer_size);

DFP_STDCALL
DfpFormatTypeNameW (DfPolicyVar var, wchar_t* buffer, size_t buffer_size);

#ifdef UNICODE
#define DfpFormatTypeName DfpFormatTypeNameW
#else
#define DfpFormatTypeName DfpFormatTypeNameA
#endif /* UNICODE */

/*------------------------------------------------------------------- 
  Table operations
  -------------------------------------------------------------------*/
DFP_STDCALL
DfpGetTableDimensions (DfPolicyVar table, int* rows, int* columns);

DFP_STDCALL
DfpGetColumnType (DfPolicyVar table, int column, enum DfPolicyType*);

DFP_STDCALL
DfpInsertColumn (DfPolicyVar table, int position, enum DfPolicyType type);

DFP_STDCALL
DfpInsertRow (DfPolicyVar table, int row_position);

DFP_STDCALL
DfpRemoveRow (DfPolicyVar table, int row);

/*------------------------------------------------------------------- 
  Table cell access
  -------------------------------------------------------------------*/

DFP_STDCALL
DfpGetCellAccessRights (DfPolicyVar var, int row, int col, enum DfPolicyAccess*);

DFP_STDCALL
DfpGetCellValue (DfPolicyVar table, int row, int col, DfPolicyValue* value, size_t value_size);

DFP_STDCALL
DfpGetCellValueSize (DfPolicyVar table, int row, int col, size_t* size_ptr);

DFP_STDCALL
DfpSetCellValue (DfPolicyVar table, int row, int col, const DfPolicyValue* value);

/* Returns an integer value or DFP_ERR_NO_VALUE */
DFP_STDCALL
DfpGetCellInt (DfPolicyVar table, int row, int col, df_uint* value);

DFP_STDCALL
DfpGetCellStringA (DfPolicyVar table, int row, int col, char* value, size_t value_size);

DFP_STDCALL
DfpGetCellStringW (DfPolicyVar table, int row, int col, wchar_t* value, size_t value_size);

#ifdef UNICODE
#define DfpGetCellString DfpGetCellStringW
#else
#define DfpGetCellString DfpGetCellStringA
#endif /* UNICODE */

DFP_STDCALL
DfpSetCellInt (DfPolicyVar table, int row, int col, df_uint value);

DFP_STDCALL
DfpSetCellStringA (DfPolicyVar table, int row, int col, const char* value);

DFP_STDCALL
DfpSetCellStringW (DfPolicyVar table, int row, int col, const wchar_t* value);

#ifdef UNICODE
#define DfpSetCellString DfpSetCellStringW
#else
#define DfpSetCellString DfpSetCellStringA
#endif /* UNICODE */

/*------------------------------------------------------------------- 
  Restriction operations
  -------------------------------------------------------------------*/
#define DFP_HIDDEN				1
#define DFP_FINAL				2
#define DFP_CHOICE				4
#define DFP_RANGE				8
#define DFP_RESTRICTION_MASK	0x0F
#define DFP_FIXED_SIZE			0x10

struct DfPolicyRestriction__ { int unused; };
typedef struct DfPolicyRestriction__* DfPolicyRestriction;

/* Return a combination of the above constants */
DFP_STDCALL
DfpGetRestriction (DfPolicyVar var, 
				   unsigned int* rtype, DfPolicyRestriction* restr);

DFP_STDCALL
DfpGetCellRestriction (DfPolicyVar table, 
					   int row, int col,
					   unsigned int* rtype, 
					   DfPolicyRestriction* restr);

typedef struct {
    int is_signed;
    union {
		struct {
			int minimum;
			int maximum;
		} s_bounds;
		struct {
			unsigned int minimum;
			unsigned int maximum;
		} u_bounds;
    } data;
} DfPolicyRange;

DFP_STDCALL
DfpGetRange (DfPolicyRestriction restr, DfPolicyRange* range);

DFP_CALL(int)
DfpGetNumChoices (DfPolicyRestriction restr);

DFP_STDCALL
DfpGetChoice (DfPolicyRestriction restr, int index, 
			  DfPolicyValue* value, size_t value_size);

DFP_STDCALL
DfpGetChoiceInt (DfPolicyRestriction restr, int index, 
			     df_uint* choice_value);
DFP_STDCALL
DfpGetChoiceStringA (DfPolicyRestriction restr, int index, 
					 char* buffer, size_t buffer_size);
DFP_STDCALL
DfpGetChoiceStringW (DfPolicyRestriction restr, int index, 
					 wchar_t* buffer, size_t buffer_size);

#ifdef UNICODE
#define DfpGetChoiceString DfpGetChoiceStringW
#else
#define DfpGetChoiceString DfpGetChoiceStringA
#endif /* UNICODE */

#ifdef __cplusplus
}
#endif

#endif /* POLICY_H */
